from enum import StrEnum

from pydantic import Field, field_validator
from pydantic_settings import BaseSettings


class Action(StrEnum):
    APPLY = "apply"
    DESTROY = "destroy"


class EnvVar:
    ACTION = "ACTION"
    DRY_RUN = "DRY_RUN"
    LOG_LEVEL = "LOG_LEVEL"
    INPUT_FILE = "INPUT_FILE"
    BACKEND_TF_FILE = "BACKEND_TF_FILE"
    OUTPUTS_FILE = "OUTPUTS_FILE"
    PLAN_FILE_JSON = "PLAN_FILE_JSON"
    TERRAFORM_CMD = "TERRAFORM_CMD"
    TF_VARS_FILE = "TF_VARS_FILE"
    VARIABLES_TF_FILE = "VARIABLES_TF_FILE"


class Config(BaseSettings):
    """Environment Variables."""

    # general settings
    action: Action = Field(Action.APPLY, alias=EnvVar.ACTION)
    dry_run: bool = Field(default=True, alias=EnvVar.DRY_RUN)
    log_level: str = Field("INFO", alias=EnvVar.LOG_LEVEL)

    # app-interface input related
    input_file: str = Field("/inputs/input.json", alias=EnvVar.INPUT_FILE)

    backend_tf_file: str = Field("module/backend.tf", alias=EnvVar.BACKEND_TF_FILE)
    outputs_file: str = Field("tmp/outputs.json", alias=EnvVar.OUTPUTS_FILE)
    plan_file_json: str = Field("tmp/plan.json", alias=EnvVar.PLAN_FILE_JSON)
    terraform_cmd: str = Field("terraform", alias=EnvVar.TERRAFORM_CMD)
    tf_vars_file: str = Field("module/terraform.tfvars.json", alias=EnvVar.TF_VARS_FILE)
    variables_tf_file: str = Field(
        "module/variables.tf", alias=EnvVar.VARIABLES_TF_FILE
    )

    @field_validator("action", mode="before")
    @classmethod
    def action_lower(cls, v: str) -> str:
        """Always lower action string to match with Action enum."""
        return v.lower()



from .generators import (
    create_backend_tf_file,
    create_tf_vars_json,
    create_variables_tf_file,
)
from .plan import (
    Action,
    Change,
    DeferredResourceChange,
    Plan,
    ResourceAttribute,
    ResourceChange,
    TerraformJsonPlanParser,
)
from .run import terraform_run

__all__ = [
    "Action",
    "Change",
    "DeferredResourceChange",
    "Plan",
    "ResourceAttribute",
    "ResourceChange",
    "TerraformJsonPlanParser",
    "create_backend_tf_file",
    "create_tf_vars_json",
    "create_variables_tf_file",
    "terraform_run",
]



import base64
import json
import os
from collections.abc import Mapping
from pathlib import Path
from typing import Any, TypeVar

from pydantic import BaseModel

from external_resources_io.config import Config


class TerraformProvisionOptions(BaseModel):
    tf_state_bucket: str
    tf_state_region: str
    tf_state_dynamodb_table: str
    tf_state_key: str


class AppInterfaceProvision(BaseModel):
    provision_provider: str  # aws
    provisioner: str  # ter-int-dev
    provider: str  # aws-iam-role
    identifier: str
    target_cluster: str
    target_namespace: str
    target_secret_name: str | None
    module_provision_data: TerraformProvisionOptions


T = TypeVar("T", bound=BaseModel)


def parse_model(model_class: type[T], data: Mapping[str, Any]) -> T:
    return model_class.model_validate(data)


def read_input_from_file(file_path: Path | str | None = None) -> dict[str, Any]:
    return json.loads(
        Path(file_path or Config().input_file).read_text(encoding="utf-8")
    )


def read_input_from_env_var(var: str = "INPUT") -> dict[str, Any]:
    b64data = os.environ[var]
    str_input = base64.b64decode(b64data.encode("utf-8")).decode("utf-8")
    return json.loads(str_input)


def get_ai_provision_data() -> AppInterfaceProvision:
    """Get the AppInterfaceProvision from the input data file."""
    ai_input = read_input_from_file()
    return parse_model(AppInterfaceProvision, ai_input["provision"])



import logging
import subprocess
from collections.abc import Sequence

from external_resources_io.config import Config

logger = logging.getLogger(__name__)


def terraform_available() -> bool:
    try:
        subprocess.run(["terraform", "--version"], check=True, capture_output=True)
        return True
    except (subprocess.CalledProcessError, FileNotFoundError):
        return False


def terraform_fmt(data: str) -> str:
    if not terraform_available():
        return data
    return subprocess.run(
        ["terraform", "fmt", "-"],
        input=data,
        text=True,
        check=True,
        capture_output=True,
    ).stdout


def terraform_run(args: Sequence[str], *, dry_run: bool | None = None) -> str:
    """Run a terraform command."""
    config = Config()
    args = [*config.terraform_cmd.split(), *args]
    dry_run = dry_run if dry_run is not None else config.dry_run
    if dry_run:
        logger.debug(f"cmd: {' '.join(args)}")
        return ""

    try:
        cmd = subprocess.run(args, capture_output=True, text=True, check=True)
    except subprocess.CalledProcessError as e:
        logger.exception(e.stdout)
        logger.exception(e.stderr)
        raise
    return cmd.stdout



import logging
import logging.config

from external_resources_io.config import Config


class DryRunFilter(logging.Filter):
    """Adds a DRY_RUN prefix"""

    def __init__(self, *, dry_run: bool) -> None:
        super().__init__()
        if dry_run:
            self.prefix = "DRY_RUN - "
        else:
            self.prefix = ""

    def filter(self, record: logging.LogRecord) -> bool:
        """Filter method"""
        record.prefix = self.prefix
        return True


def setup_logging() -> None:
    """Returns a logger"""
    config = Config()
    logging.config.dictConfig({
        "version": 1,
        "disable_existing_loggers": False,
        "filters": {"prefix_filter": {"()": DryRunFilter, "dry_run": config.dry_run}},
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "filters": ["prefix_filter"],
                "formatter": "base",
            }
        },
        "formatters": {"base": {"format": "%(prefix)s%(levelname)s - %(message)s"}},
        "root": {"level": config.log_level, "handlers": ["console"]},
        "botocore": {"level": "ERROR", "handlers": ["console"]},
    })



import importlib
from pathlib import Path
from typing import Annotated, Protocol, cast

from pydantic import BaseModel

from external_resources_io.config import Config, EnvVar
from external_resources_io.input import (
    AppInterfaceProvision,
    parse_model,
    read_input_from_file,
)
from external_resources_io.terraform.generators import (
    create_backend_tf_file,
    create_tf_vars_json,
    create_variables_tf_file,
)

try:
    import typer
except ImportError:
    raise ImportError(
        "Please install external-resources-io with `pip install external-resources-io[cli]`"
    ) from None


app = typer.Typer()
tf_app = typer.Typer()
app.add_typer(tf_app, name="tf")
config = Config()


class AppInterfaceInputInterface(Protocol):
    data: BaseModel
    provision: AppInterfaceProvision


def _get_app_interface_class(app_interface_input_class: str) -> type[BaseModel]:
    ai_module_name, ai_class_name = app_interface_input_class.rsplit(".", maxsplit=1)
    ai_class = getattr(importlib.import_module(ai_module_name), ai_class_name)
    if not issubclass(ai_class, BaseModel):
        raise TypeError(
            f"{app_interface_input_class} must be a subclass of pydantic.BaseModel"
        )
    return ai_class


def _get_app_interface_data_class(app_interface_input_class: str) -> type[BaseModel]:
    data_class = (
        _get_app_interface_class(app_interface_input_class)
        .model_fields["data"]
        .annotation
    )
    if not data_class or not issubclass(data_class, BaseModel):
        raise TypeError(f"{data_class} must be a subclass of pydantic.BaseModel")
    return data_class


def _get_ai_input(
    app_interface_input_class: str, input_file: Path | None
) -> AppInterfaceInputInterface:
    """Get the AppInterfaceInput from the input file."""
    return cast(
        "AppInterfaceInputInterface",
        parse_model(
            _get_app_interface_class(app_interface_input_class),
            read_input_from_file(input_file),
        ),
    )


@tf_app.command()
def generate_variables_tf(
    app_interface_input_class: Annotated[
        str,
        typer.Argument(
            help="App interface input class. E.g. your_module_name.input.AppInterfaceInput",
            show_default=False,
        ),
    ],
    output: Annotated[
        Path | None,
        typer.Option(
            help="Output file",
            dir_okay=False,
            writable=True,
            envvar=EnvVar.OUTPUTS_FILE,
        ),
    ] = Path(config.variables_tf_file),
) -> None:
    """Generates Terraform variables.tf file."""
    create_variables_tf_file(
        _get_app_interface_data_class(app_interface_input_class), output
    )


@tf_app.command()
def generate_backend_tf(
    app_interface_input_class: Annotated[
        str,
        typer.Argument(
            help="App interface input class. E.g. your_module_name.input.AppInterfaceInput",
            show_default=False,
        ),
    ],
    input_file: Annotated[
        Path,
        typer.Argument(
            help="The App interface input json file",
            show_default=False,
            readable=True,
            file_okay=True,
            envvar=EnvVar.INPUT_FILE,
        ),
    ],
    output: Annotated[
        Path | None,
        typer.Option(
            help="Output file",
            dir_okay=False,
            writable=True,
            envvar=EnvVar.OUTPUTS_FILE,
        ),
    ] = Path(config.backend_tf_file),
) -> None:
    """Generates Terraform backends.tf file."""
    ai_input = _get_ai_input(app_interface_input_class, input_file)
    create_backend_tf_file(ai_input.provision, output)


@tf_app.command()
def generate_tf_vars_json(
    app_interface_input_class: Annotated[
        str,
        typer.Argument(
            help="App interface input class. E.g. your_module_name.input.AppInterfaceInput",
            show_default=False,
        ),
    ],
    input_file: Annotated[
        Path,
        typer.Argument(
            help="The App interface input json file",
            show_default=False,
            readable=True,
            file_okay=True,
            envvar=EnvVar.INPUT_FILE,
        ),
    ],
    output: Annotated[
        Path | None,
        typer.Option(
            help="Output file",
            dir_okay=False,
            writable=True,
            envvar=EnvVar.OUTPUTS_FILE,
        ),
    ] = Path(config.tf_vars_file),
) -> None:
    """Generates Terraform tfvars.json file."""
    ai_input = _get_ai_input(app_interface_input_class, input_file)
    create_tf_vars_json(ai_input.data, output)


if __name__ == "__main__":
    app()



# ruff: noqa: ANN401
import json
from collections.abc import Sequence
from pathlib import Path
from types import UnionType
from typing import Any, Literal, Union, get_args, get_origin

from pydantic import BaseModel
from pydantic_core import PydanticUndefined

from external_resources_io.config import Config
from external_resources_io.input import AppInterfaceProvision
from external_resources_io.terraform.run import terraform_fmt


class SetEncoder(json.JSONEncoder):
    def default(self, obj: Any) -> Any:
        if isinstance(obj, set):
            return list(obj)
        return json.JSONEncoder.default(self, obj)


def create_tf_vars_json(
    input_data: BaseModel,
    output_file: Path | str | None = None,
    *,
    exclude_none: bool = True,
) -> Path:
    """Helper method to create teraform vars files. Used in terraform based ERv2 modules."""
    output = Path(output_file or Config().tf_vars_file)
    output.write_text(
        input_data.model_dump_json(exclude_none=exclude_none),
        encoding="utf-8",
    )
    return output


def create_backend_tf_file(
    provision_data: AppInterfaceProvision, output_file: Path | str | None = None
) -> Path:
    """Helper method to create teraform backend configuration. Used in terraform based ERv2 modules."""
    output = Path(output_file or Config().backend_tf_file)
    output.write_text(
        terraform_fmt(f"""\
            terraform {{
              backend "s3" {{
                bucket = "{provision_data.module_provision_data.tf_state_bucket}"
                key    = "{provision_data.module_provision_data.tf_state_key}"
                region = "{provision_data.module_provision_data.tf_state_region}"
                dynamodb_table = "{provision_data.module_provision_data.tf_state_dynamodb_table}"
                profile = "external-resources-state"
              }}
            }}"""),
        encoding="utf-8",
    )
    return output


def create_variables_tf_file(
    model: type[BaseModel], variables_file: Path | str | None = None
) -> Path:
    """Generates Terraform variables.tf file."""
    output = Path(variables_file or Config().variables_tf_file)
    output.write_text(
        terraform_fmt(
            _convert_json_to_hcl(_generate_terraform_variables_from_model(model))
        ),
        encoding="utf-8",
    )
    return output


def _generate_fields(model: type[BaseModel]) -> dict[str, dict]:
    return {
        field_name: _generate_terraform_variable(
            python_type=field_info.annotation, default=field_info.default
        )
        for field_name, field_info in model.model_fields.items()
    }


def _generate_terraform_variables_from_model(model: type[BaseModel]) -> dict:
    """Generates Terraform variables json."""
    return {"variable": _generate_fields(model)}


def _generate_terraform_variable(python_type: Any, default: Any = None) -> dict:
    """Generates a Terraform variable block."""
    variable_block: dict[str, str | None] = {"type": _get_terraform_type(python_type)}

    if default is not PydanticUndefined:
        variable_block["default"] = (
            default.model_dump()
            if default and isinstance(default, BaseModel)
            else default
        )
    return variable_block


def _convert_generic_types(origin: Any, args: Any) -> str:  # noqa: PLR0911
    """Convert generic types to Terraform types."""
    match origin:
        case t if t in {list, Sequence}:
            return f"list({_get_terraform_type(args[0])})" if args else "list(any)"

        case t if t is set:
            return f"set({_get_terraform_type(args[0])})" if args else "set(any)"

        case t if t is dict:
            return f"map({_get_terraform_type(args[1])})" if args else "map(any)"

        case t if t is Literal:
            return "string"

        case t if t in {UnionType, Union}:
            if type(None) in args:
                return _get_terraform_type(args[0])
            return "any"

        case _:
            return "any"


def _convert_basic_types(python_type: Any) -> str:  # noqa: PLR0911
    """Convert basic python types to Terraform types."""
    match python_type:
        case t if t is str:
            return "string"
        case t if t is int:
            return "number"
        case t if t is bool:
            return "bool"
        case t if t is list:
            return "list(any)"
        case t if t is set:
            return "set(any)"
        case t if t is dict:
            return "map(any)"
        case t if issubclass(t, BaseModel):
            # nested model
            fields_types = ",".join(
                f"{k} = {v['type']}" for k, v in _generate_fields(t).items()
            )
            return f"object({{{fields_types}}})"
        case _:
            return "any"


def _get_terraform_type(python_type: Any) -> str:
    """Maps Python types to Terraform types."""
    origin = get_origin(python_type)
    args = get_args(python_type)

    return (
        _convert_generic_types(origin, args)
        if origin is not None
        else _convert_basic_types(python_type)
    )


def _convert_json_value_to_hcl(value: Any) -> str:  # noqa: PLR0911
    """Converts a JSON value to HCL."""
    match value:
        case t if isinstance(t, str):
            return f'"{value}"'
        case t if isinstance(t, bool):
            return str(value).lower()
        case t if isinstance(t, int | float):
            return str(value)
        case t if isinstance(t, list | set):
            if not value:
                return "[]"
            return "[" + ",".join(_convert_json_value_to_hcl(e) for e in value) + "]"
        case t if isinstance(t, dict):
            if not value:
                return "{}"
            pairs = []
            for k, v in value.items():
                converted_v = _convert_json_value_to_hcl(v)
                pairs.append(f"{k} = {converted_v}")
            return "{\n" + "\n".join(pairs) + "\n}"
        case None:
            return "null"
        case _:
            return str(value)


def _convert_json_to_hcl(data: dict) -> str:
    variables = data.get("variable", {})
    hcl_blocks = []

    for var_name, var_config in sorted(variables.items()):
        block_lines = [f'variable "{var_name}" {{']
        for key, value in var_config.items():
            hcl_value = value if key == "type" else _convert_json_value_to_hcl(value)
            block_lines.append(f"{key} = {hcl_value}")
        block_lines.append("}\n")
        hcl_blocks.append("\n".join(block_lines))

    return "\n".join(hcl_blocks)
